var Graph = (function () {
    'use strict';
    /**
     * @description The main class, she uses Cytoscape.js for the graph visualization, ask ElementJSON the nodes coordinates when using a Viz.js layout.
     * @param {ElementJSON} element_json
     * @typedef Graph Manage the graph visualization.
     */
    function Graph(element_json) {
        /**
         * @description Set all the cytoscape based layout settings.
         * @type object
         */
        var settings = {
            /* Cola.js settings. */
            cola: {
                name: "cola",
                nodeSpacing: function (node) { return 15; },
                padding: 30,
                flow : {axis: 'x', minSeparation: 400},
                animate: true,
                maxSimulationTime: 2000
            },
            /* BREADTHFIRST settings. */
            breadthfirst: {
                name: "breadthfirst",
                padding: 30,
                animate: true,
                animationDuration: 650,
                roots: [],
                directed: true
            },
            /* Common settings (GRID, CIRCLE, CONCENTRIC) */
            other: {
                name: "",
                padding: 30,
                animate: true,
                animationDuration: 650
            }
        };
        /**
         * @description Special settings for Cola.js when getting positions from a Viz.js layout.
         * @type {object}
         */
        var cola = {
            name: "cola",
            padding: 30,
            nodeSpacing: 15,
            animate: true,
            maxSimulationTime: 600
        };
        /**
         * @description Cytoscape styles.
         * @type object
         */
        var style_cytoscape;
        /**
         * @description Event manager
         * @type GraphEventManager
         */
        var _eventManager;
        /**
         * @description The cytoscape element.
         * @type object
         */
        var cy;
        /**
         * @description Contains the cytoscape nodes differenciate in groups and simple nodes.
         * @type {{groups: object, simple: object}}
         */
        var _nodes = {
            groups : [],
            simple : []
        };
        /**
         * @description Cytoscape layout.
         * @type object
         */
        var cy_layout;
        /**
         * @description Last wanted layout:
         *      - type, the type of layout (cytoscape or viz);
         *      - name, the name of the layout (grid, dot, etc...);
         *      - is_running, if the layout is currently running.
         * @type {{type:string, name:string, is_running:boolean}}
         */
        var layout = {type:"", name:"", is_running:false};
        /**
         * @description Used by autoCollapse function to collapse or expand the groups of nodes.
         * @type {boolean}
         */
        var is_collapse = (typeof(Storage) && sessionStorage.getItem("collapse_group") ? sessionStorage.getItem("collapse_group") : true);

        /**
         * @description Give the cytoscape instance.
         * @returns {object|Graph.cy}
         */
        this.getCy = function () { return cy; };
        /**
         * @description Get the node group. Not the grouped nodes, the parent node.
         * @returns {Object}
         */
        this.groups = function () {
            return _nodes.groups;
        };
        /**
         * @description Get the nodes without the node group.
         * @returns {Object}
         */
        this.nodes = function () {
            return _nodes.simple;
        };
        /**
         * @description Set how to show the groups collapse or extend. Collapsed groups will hide the nodes they contains.
         * @param {string|boolean} collapse Give 'true' or 'false'. Parse to boolean inside.
         */
        this.setCollapse = function (collapse) {
            collapse = collapse === "true";
            if (collapse !== is_collapse) {
                is_collapse = collapse;
                autoCollapse();
            }
        };
        /**
         * @description Instantiate a new Cytoscape element with new json elements.
         * @param {object[]} json
         */
        function setCyElement(json, layout) {
            /* Turn off the event on the nodes and destroying this cytoscape instance. */
            cy.nodes().off('select');
            cy.nodes().off('unselect');
            cy.off('layoutstop');
            cy.destroy();
            /* Create a new cytoscape with the new elements. */
            cy = cytoscape({
                container: $("#cy"),

                elements: json,

                style: style_cytoscape,

                layout: layout,

                maxZoom: 6
            });

            cy.on('layoutstop', function () {
                cy.navigator("destroy");
                cy.navigator();
                cy.panzoom({
                    zoomOnly    : true,
                    fitPadding  : 30,
                    maxZoom     : 6
                });

                autoCollapse();

                layout.is_running = true;
            });

            if (cy.nodes().length > 0) {
                if (!$("#allnodefilter").hasClass("hidden")) $("#allnodefilter").addClass("hidden");
                /* Prepare the groups and collapse the edges. */
                cy.batch(function () {
                    _nodes.groups = cy.nodes("$node > node");
                    _nodes.simple = cy.nodes().difference(_nodes.groups);

                    /* All groups will know their cytoscape child. */
                    _nodes.groups.forEach(function (node) { node.data(node_a.data.hiddenProperties)[node_p.me].setProp(node_p.cytoscapeNodes, node.children()); });


                    /* Collapse all filter edge. */
                    var edge_to_collapse = {};
                    cy.edges(".filter").codirectedEdges(".filter").forEach(function (edge) {
                        var edgeToCollapseId = edge.data(edge_a.data.source)+"_linkbetweensource_to_target_" + edge.data(edge_a.data.target);
                        if (!edge_to_collapse[edgeToCollapseId]) edge_to_collapse[edgeToCollapseId] = [edge];
                        else                                     edge_to_collapse[edgeToCollapseId].push(edge);
                    });

                    for (var key in edge_to_collapse) {
                        var arrayOfEdge = edge_to_collapse[key];
                        if (arrayOfEdge.length > 1) {
                            var nodes = [];
                            for (var e = 0, edge = arrayOfEdge[e] ; e < arrayOfEdge.length ; e++, edge = arrayOfEdge[e]) {
                                nodes = nodes.concat(edge.data(edge_a.data.hiddenProperties)[edge_p.me].getProp(edge_p.nodes));
                                edge.remove();
                            }
                            var sourceTarget = key.split("_linkbetweensource_to_target_");
                            var edgeCollapse = { data: { source: sourceTarget[0], target: sourceTarget[1] }, group: elementGroup.edge};
                            var edge = new GEdge(edgeCollapse);
                            edge.setProp(edge_p.source, cy.nodes("#" + edge.data(edge_a.data.source)).data(node_a.data.hiddenProperties)[node_p.me]);
                            edge.setProp(edge_p.target, cy.nodes("#" + edge.data(edge_a.data.target)).data(node_a.data.hiddenProperties)[node_p.me]);
                            edge.setProp(edge_p.nodes, nodes.filter(function (value, index, array) {
                                return array.lastIndexOf(value) === index;
                            }));
                            cy.add(edge.basicElement());
                        }
                    }
                });
                /* Supplementary initializations. */
                initLabelNodes();
                initBreadthfirstRoot();
                initColaTime();

                _eventManager.initEventCy(cy);

                /* Select the last selected nodes. */
                var select = false;
                cy.nodes().forEach(function (node) {
                    var json_node = node.data(node_a.data.hiddenProperties)[node_p.me];
                    if (json_node.areYou(node_p.selected)) {
                        select = true;
                        json_node.unselect();
                        node.select();
                    }
                });

                if (!select) majGraphColor();
            } else {
                if ($("#allnodefilter").hasClass("hidden")) $("#allnodefilter").removeClass("hidden");
            }
        }

        /**
         * @description Set up the nodes label to show to the user.
         */
        function initLabelNodes() {
            cy.batch(function () {
                _nodes.simple.forEach(function (node) {
                    node.style("label", node.data(node_a.data.description) + '\n' + node.data(node_a.data.class));
                });
            });
        }
        /**
         * @description Set up the breadthfirst roots elements.
         */
        function initBreadthfirstRoot() { settings.breadthfirst.roots = cy.nodes().roots(); }
        /**
         * @description Set the maxSimulationTime for the cola layout.
         */
        function initColaTime() {
            var nbrElement = cy.nodes().size();
            settings.cola.maxSimulationTime = nbrElement * 10;
        }

        /**
         * @description Update the node.
         * @param {object} node A cytoscape node.
         * @param {boolean} selection Whether or not the update belongs to a selection or a deselection.
         * @public
         */
        this.majNode = function (node, selection) {
            majNode(node.data(node_a.data.hiddenProperties)[node_p.me], selection);
            majGraphColor();
        };
        /**
         * @description Update a group of node.
         * @param {object} group A cytoscape node.
         * @param {boolean} selection Whether or not the update belongs to a selection or a deselection.
         */
        this.majGroup = function (group, selection) {
            var json_group = group.data(node_a.data.hiddenProperties)[node_p.me];
            var nodes = json_group.getProp(node_p.cytoscapeNodes);

            for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) {
                if (node.inside())  {
                    if (selection) {
                        if (!node.selected()) node.select();
                    } else {
                        if (node.selected())  node.unselect();
                    }
                } else {
                    if (selection || !selection && node.data(node_a.data.hiddenProperties)[node_p.me].areYou(node_p.selected))
                        majNode(node.data(node_a.data.hiddenProperties)[node_p.me], selection);
                }
            }
            majGraphColor();
        };
        /**
         * @description Update a group of node.
         * @param {object} group A cytoscape node.
         * @param {boolean} selection Whether or not the update belongs to a selection or a deselection.
         */
        function majGroup (group, selection) {
            var json_group = group.data(node_a.data.hiddenProperties)[node_p.me];
            var nodes = json_group.getProp(node_p.cytoscapeNodes);

            for (var n = 0, node = nodes[n]; n < nodes.length; n++, node = nodes[n]) {
                if (node.inside())  {
                    if (selection) {
                        if (!node.selected()) node.select();
                    } else {
                        if (node.selected())  node.unselect();
                    }
                } else {
                    majNode(node.data(node_a.data.hiddenProperties)[node_p.me], selection);
                }
            }
            majGraphColor();
        }
        /**
         * @description Update the node.
         * @param {GNode} node A cytoscape node.
         * @param {boolean} selection Whether or not the update belongs to a selection or a deselection.
         * @private
         */
        function majNode(node, selection) { element_json.majNode(node, selection); }

        /**
         * @description Tell if this node is root.
         * @param {object} node
         * @returns {boolean}
         */
        this.estRoot = function (node) { return node.anySame(cy.nodes().roots()); };
        /**
         * @description Graph resizing. If nothing is specified, fit on the graph with a padding of 30.
         * @param {string} [type] "center" | "fit"
         * @param {string} [selected] "true" | "false"
         */
        this.resize = function (type, selected) {
            if (typeof type === "undefined") {
                cy.fit({}, 30);
            } else {
                var selection = {};
                if (selected === "true") {
                    selection = cy.elements(":selected");
                }
                switch (type) {
                    case "fit":
                        cy.fit(selection, 30);
                        break;
                    case "center":
                        cy.center(selection);
                        break;
                }
            }
        };

        /**
         * @description Call the collapse or the uncollapse function.
         */
        function autoCollapse () {
            if (is_collapse)     collapse_all();
            else                expand_all();

            majGraphColor();
        }
        /**
         * @description All parent nodes will be resized and their children will be remove from the graph.
         */
        function collapse_all () {
            var busy =false, task = "collapse_group", group_iterator = new IteratorGraph(_nodes.groups);

            if (group_iterator.size() > 0) {
                $("#collapsing").removeClass("hidden");

                group_iterator.beforeFirst();
                var interval = setInterval(function () {
                    if (!busy) {
                        busy = !busy;

                        switch (task) {
                            case "collapse_group":
                                if (group_iterator.hasNext()) {
                                    if (!group_iterator.next().hasClass("collapse_group")) {
                                        collapse_element(group_iterator.current());
                                    }
                                } else {
                                    task = "finish";
                                }
                                break;
                            case "finish":
                                $("#collapsing").addClass("hidden");
                                clearInterval(interval);
                                break;
                        }

                        busy = !busy;
                    }
                }, 100);
            }
        }
        /**
         * @description Collapse a specific element.
         * @param {object} node
         */
        function collapse_element (node) {
            var node_iterator = new IteratorGraph(node.children());

            node_iterator.forEach(function (child) {
                if (child.selected()) child.unselect();

                var incomersEdges = child.incomers("edge"),
                    outgoersEdges = child.outgoers("edge");

                moveEdges(incomersEdges, child.id(), edge_a.data.target, {target: node.id()});
                moveEdges(outgoersEdges, child.id(), edge_a.data.source, {source: node.id()});

                removeAllClasses(child);
                child.remove()
            });

            /* Adding collapse class to the node group. */
            node.addClass("collapse_group");
            node.removeClass("expand_group");

            if (node.selected()) {
                majGroup(node, true);
            } else {
                majNodeColor(node);
            }
            /**
             * @description Moving the given edges and storing their original location.
             * @param {object} edges A collection of cytoscape edges.
             * @param {string} originalId
             * @param {string} linkWay
             * @param {{source:string}|{target:string}} newLocation
             */
            function moveEdges (edges, originalId, linkWay, newLocation) {
                for (var e = 0, edge = edges[e]; e < edges.length; e++, edge = edges[e]) {
                    if (linkWay === edge_a.data.source) edge.data(edge_a.data.realSource, originalId);
                    else                                edge.data(edge_a.data.realTarget, originalId);

                    edge.move(newLocation);
                }
            }
        }
        /**
         * @description Used by the context menu to allow users to specifically collapse a group node.
         * @param {object} node
         */
        this.collapse_element = function (node) { collapse_element(node); };
        /**
         * @description Expand all group.
         */
        function expand_all () {
            var busy = false, task = "expand_group", group_iterator = new IteratorGraph(_nodes.groups);

            if (group_iterator.size() > 0) {
                $("#expanding").removeClass("hidden");

                group_iterator.beforeFirst();
                var interval = setInterval(function () {
                    if (!busy) {
                        busy = !busy;

                        switch (task) {
                            case "expand_group":
                                if (group_iterator.hasNext()) {
                                    if (!group_iterator.next().hasClass("expand_group")) {
                                        expand_element(group_iterator.current());
                                    }
                                } else {
                                    task = "finish";
                                }
                                break;
                            case "finish":
                                clearInterval(interval);
                                $("#expanding").addClass("hidden");
                                break;
                        }

                        busy = !busy;
                    }
                }, 100);
            }
        }
        /**
         * @description Expand a specific element.
         * @param {object} node A Cytoscape node.
         */
        function expand_element (node) {
            /* Resizing the node group. */
            var json_node = node.data(node_a.data.hiddenProperties)[node_p.me];

            var node_iterator = new IteratorGraph(json_node.getProp(node_p.cytoscapeNodes));

            node_iterator.forEach(function (child) {
                child.restore();

                /* Moving the edges to their original target/source. */
                node.incomers("edge[" + edge_a.data.realTarget + "='" + child.id() + "']").forEach(function (edge) { edge.move({target : child.id()}); });
                node.outgoers("edge[" + edge_a.data.realSource + "='" + child.id() + "']").forEach(function (edge) { edge.move({source : child.id()}); });

                majNodeColor(child);
            });

            /* Removing collapse class to the node group. */
            node.addClass("expand_group");
            node.removeClass("collapse_group");
            majNodeColor(node);
        }
        /**
         * @description Used by the context menu to allow users to specifically expand a group node.
         * @param {object} node
         */
        this.expand_element = function (node) { expand_element(node); };

        /**
         * @description Graph color update
         * @public
         */
        this.majGraphColor = function () {
            majGraphColor();
        };
        /**
         * @description Graph color update
         * @private
         */
        function majGraphColor() {
            cy.batch(function () {
                majNodesColor();
                majEdgesColor();
            });
        }
        /**
         * @description Reset the color class of this element.
         * @param {object} element
         */
        function removeAllClasses(element) {
            element.removeClass(css_classes_array.join(" "));
        }
        /**
         * @description Update the nodes classes.
         */
        function majNodesColor() {
            cy.nodes().forEach(majNodeColor);
        }
        /**
         * @description Update this node classes.
         * @param node
         */
        function majNodeColor(node) {
            removeAllClasses(node);
            if (!node.hasClass("expand_group")) {
                var json_node = node.data(node_a.data.hiddenProperties)[node_p.me];
                node.addClass(json_node.getClasses().join(" "));
            }
        }
        /**
         * @description Update the edges classes.
         */
        function majEdgesColor() {
            cy.edges().forEach(majEdgeColor);
        }
        /**
         * @description Update this edge classes.
         * @param edge
         */
        function majEdgeColor (edge) {
            removeAllClasses(edge);
            var json_edge = edge.data(edge_a.data.hiddenProperties)[edge_p.me];
            if (json_edge.areYou(edge_p.nodes)) {
                var nodes = json_edge.getProp(edge_p.nodes);
                var classArr = [];
                for (var n = 0, node = nodes[n]; n < nodes.length ; n++, node = nodes[n]) {
                    if (node.areYou(node_p.impacted)) classArr.push(node.data(node_a.data.impactNode));
                }

                if (classArr.length > 0) {
                    edge.addClass(getClassColor(Math.min.apply(null, classArr)));
                    edge.addClass(css_classes.impacted);
                }
            } else {
                edge.addClass(json_edge.source().getClasses().join(" "));
            }
        }

        /**
         * @description Launch the layout only if the layout is not running.
         */
        this.run = function () {
            if (!layout.is_running) {
                if (layout.type === "cyto") {
                    runCyto();
                } else if (layout.type === "viz") {
                    runViz();
                } else {
                    throw "Layout inconnu.";
                }
            }
        };
        /**
         * @description Launch a Cytoscape.js layout.
         */
        function runCyto() {
            var paramToUse = "";
            switch (layout.name) {
                case settings.cola.name:
                    paramToUse = layout.name;
                    break;
                case settings.breadthfirst.name:
                    paramToUse = layout.name;
                    break;
                default:
                    paramToUse = "other";
                    settings[paramToUse].name = layout.name;
                    break;
            }

            var doLayout = true, element = element_json.getElementToUse(), numberOfNodes = element_json.getNumberOfNodesForJsonToUse();
            if (numberOfNodes > sizeToConfirm) doLayout = popup_confirm(numberOfNodes);

            if (doLayout) {
                setCyElement(element, settings[paramToUse]);
            }
        }
        /**
         * @description Launch a Viz.js layout.
         */
        function runViz() {
            $("#cy").html("");
            $("#cy").addClass("computing");

            setTimeout(function () {
                var elements = element_json.runViz();

                if (elements.ok) {
                    setCyElement(elements.element, cola);
                    $("#cy").removeClass("computing");
                }
            }, 150);
        }
        /**
         * @description Change the layout.
         * @param {{nom:string,type:string}} layout_selected
         */
        this.changeLayout = function (layout_selected) {
            if (layout_selected.type === "cyto") {
                changeLayoutCyto(layout_selected);
            } else {
                changeLayoutViz(layout_selected);
            }
        };
        /**
         * @description Change for a Cytoscape.js layout.
         * @param {{nom:string,type:string}} layout_selected
         */
        function changeLayoutCyto(layout_selected) {
            if (layout_selected.type !== layout.type || layout_selected.nom !== layout.name) {
                layout.type = layout_selected.type;
                layout.name = layout_selected.nom;
                layout.is_running = false;
            }
        }
        /**
         * @description Change for a Viz.js layout.
         * @param {{nom:string,type:string}} layout_selected
         */
        function changeLayoutViz(layout_selected) {
            var viz_layout = element_json.getVizLayout();

            if (layout_selected.nom !== viz_layout || layout.type !== layout_selected.type) {
                if (layout_selected.nom !== viz_layout) element_json.changeVizEngine(layout_selected.nom);
                layout.type = layout_selected.type;
                layout.name = layout_selected.nom;
                layout.is_running = false;
            }
        }

        /**
         * @description Group the element like the user wanted.
         * @param {string} wanted_group Utilisation usage.
         * @see ElementJSON
         */
        this.changeElement = function (wanted_group) {
            if (wanted_group !== element_json.getUsed()) {
                element_json.getElementFor(wanted_group);
                layout.is_running = false;
            }
        };

        /**
         * @description Apply a filter on the element in ElementJSON.
         * @param {string[]} valArray
         */
        this.startFilter = function (valArray) {
            element_json.getFilteredElement(valArray);

            layout.is_running = false;
        };
        /**
         * @description Remove the filter in ElementJSON.
         */
        this.endFilter = function () {
            if (element_json.isFilterJson()) {
                element_json.getUsedJson();

                layout.is_running = false;
            }
        };

        /**
         * @description Get the contextual menu for this element.
         * @param {object} evt
         * @returns {Object[]}
         */
        this.getContext = function (evt) {
            if (evt.cyTarget !== evt.cy) {
                var element = evt.cyTarget;
                if (element.isNode()) {
                    if (element.data(node_a.data.hiddenProperties)[node_p.me].areYou(node_p.nodes)) {
                        return groupContext(element);
                    } else {
                        return nodeContext(element);
                    }
                } else {
                    return edgeContext(element);
                }
            }
        };
        /**
         * @description Group contextual menu.
         * @param {object} group
         * @returns {object[]}
         */
        function groupContext (group) {
            var collapse = {
                    name: "Collapse group",
                    fun : function () {
                        collapse_element(group);
                        $(window).contextMenu('update', [
                            {
                                name: "Collapse group",
                                disable : true,
                                className : 'hidden'
                            },{
                                name: "Expand group",
                                disable: false
                            }
                        ]);
                    }
                },
                expand = {
                    name: "Expand group",
                    fun : function () {
                        expand_element(group);
                        $(window).contextMenu('update', [
                            {
                                name: "Collapse group",
                                disable : false
                            },{
                                name: "Expand group",
                                disable: true,
                                className : 'hidden'
                            }
                        ]);
                    }
                };

            if (group.hasClass("collapse_group")) {
                collapse.className = 'hidden';
                collapse.disable = true;
                delete expand.className;
                delete expand.disable;
            } else {
                expand.className = 'hidden';
                expand.disable = true;
                delete collapse.className;
                delete collapse.disable;
            }

            return [collapse, expand];
        }
        /**
         * @description Node contextual menu.
         * @param {object} node
         * @returns {object[]}
         */
        function nodeContext (node) {
            return [
                {
                    name: "Show node Info.",
                    fun : function () {
                        showNodeInfo(node);
                    }
                },{
                    name: "Filter nodes with same class",
                    fun : function () {
                        $("#input" + node.data(node_a.data.class)).prop("checked", false);
                        if (!$("#activate_classesfilter").checked) {
                            $("#activate_classesfilter").prop("checked", true);
                            $(document.querySelectorAll("input[name=\"activate_deactivate_classesfilter\"]")).change();
                        }
                        $("#graphe_action__button").click();
                    }
                },{
                    name: "Center graph on this node",
                    fun : function () {
                        var baseUrl = racineUrl + "lavoisier/" + pageName;
                        var params = {Key: "id", Value: node.id(), Filter: getLabelFiltre()};

                        var href = baseUrl + "?Key=" + params.Key + "&Value=" + params.Value +
                            ($("#causeDepthLevel").val() ? "&CauseDepth=" + $("#causeDepthLevel").val() : "" ) +
                            ($("#impactDepthLevel").val() ? "&ImpactDepth=" + $("#impactDepthLevel").val() : "") +
                            (myImpactLevel ? "&ImpactLevel="+myImpactLevel : "");
                        if (params.Filter.length > 0) href += "&Filter=" + params.Filter.join('|');
                        window.location.href = href;
                    }
                },{
                    name: "Center graph on selected nodes",
                    fun : function () {
                        var baseUrl = racineUrl + "lavoisier/" + pageName;
                        var params = {Key: "id", Value: [], Filter: getLabelFiltre()};
                        cy.nodes(":selected").forEach(function (node) { params.Value.push(node.id()); });

                        var href = baseUrl + "?Key=" + params.Key + "&Value=" + params.Value.join("|") +
                            ($("#causeDepthLevel").val() ? "&CauseDepth=" + $("#causeDepthLevel").val() : "" ) +
                            ($("#impactDepthLevel").val() ? "&ImpactDepth=" + $("#impactDepthLevel").val() : "") +
                            (myImpactLevel ? "&ImpactLevel="+myImpactLevel : "");
                        if (params.Filter.length > 0) href += "&Filter=" + params.Filter.join('|');
                        window.location.href = href;
                    }
                }
            ];
        }
        /**
         * @description Edge contextual menu.
         * @param {object} edge
         * @returns {object[]}
         */
        function edgeContext (edge) {
            return [
                {
                    name: "Show edge Info",
                    fun : function () {
                        showEdgeInfo(edge);
                    }
                }
            ];
        }

        style_cytoscape = styles_cytoscape;

        cy = cytoscape({
            elements: element_json.getElementOrigin(),

            style: style_cytoscape,

            maxZoom: 6
        });

        _nodes.groups = cy.nodes("$node > node");
        _nodes.simple = cy.nodes().difference(_nodes.groups);

        initLabelNodes();
        initBreadthfirstRoot();
        initColaTime();

        _eventManager = new GraphEventManager(this);
    }

    return Graph;
}());